/*
  AnalogReadSerial

  Reads an analog input on pin 0, prints the result to the Serial Monitor.
  Graphical representation is available using Serial Plotter (Tools > Serial Plotter menu).
  Attach the center pin of a potentiometer to pin A0, and the outside pins to +5V and ground.

  This example code is in the public domain.

  http://www.arduino.cc/en/Tutorial/AnalogReadSerial
*/

/*
   Team DUCK TAPE breathing sensor
*/

/*
   Application phases
*/
const int WAIT_FOR_READING_PHASE = 1;
const int READING_PHASE = 2;

int current_phase = WAIT_FOR_READING_PHASE;

/**
   interval to read from the belt.
*/
const int loop_delay_msecs = 1000; // measurement interval is one second

// maximum value the arduino reads from the belt (5V)
const int MAX_BELT_VALUE = 1023;

// Normal respiration rates (in and out) for an adult person at rest range from 12 to 16 breaths per minute.
const unsigned long inOutBreathingWindowInMsecs = 7 * 1000; // 7 secs
// window size we check for top values
const unsigned long topsWindowInSeconds = 60; // 60 secs

const int ASLEEP_FREQUENCY_THRESHOLD_IN_RRD_INDECES = 3 * (1000 / loop_delay_msecs);
const int SLEEPING_THRESHOLD = inOutBreathingWindowInMsecs;
const int maxBreathingRange = 300; // difference in volts between inhale and exhale in [0..1023].

// index in the RRD
int currentRRDIndex = 0;

// processing the belts data starts after the start delay
int startDelayInMsecs = 5 * 1000 * (1000 / loop_delay_msecs);

// running value for the initial pause duration
int pausedDurationInMsecs = 0;
// value we measured the previous time from the belt:
int lastMeasuredValue = MAX_BELT_VALUE;

//history size available in the round robin database
const int maxRRDsize = (100 * (1000 / loop_delay_msecs)); // size to hold 100 seconds of data
// round robin database; holds integer values
int cache[maxRRDsize];

/** 
 * values for the logic 
 */
int waitingForTrendChange = 0;
int currentTrend = 0;
int lastTrend = 0;

/**
   Debugging, simulating flags
*/
boolean debug = true;
boolean trace = false;
boolean traceTops = false;
boolean simulating = false;

/**
   The code:
*/

// the setup routine runs once when you press reset:
void setup() {
  // initialize serial communication at 9600 bits per second:
  Serial.begin(9600);

  if (debug) {
    Serial.print("RRD size in number of values: ");
    Serial.println(maxRRDsize);
  }

  // init the RRD data
  for (int k = 0; k < maxRRDsize; k++) {
    cache[k] = MAX_BELT_VALUE;
  }

  // test
  //for (int k = 0; k < 200; k++) {
  //  int startIndex = getStartIndexFromSeconds(k);
  //  if (trace) printIndex(startIndex, startIndex);
  //}
}

// the loop routine runs over and over again forever:
void loop() {

  switch (current_phase) {
    case WAIT_FOR_READING_PHASE:
      {
        // update the start delay timer
        if (pausedDurationInMsecs > startDelayInMsecs) {
          // next phase
          current_phase = READING_PHASE;
        } else {
          pausedDurationInMsecs += loop_delay_msecs;
        }
      }
      break;
    case READING_PHASE:
      {
        // read the input on analog pin 0:
        int sensorValue = analogRead(A0); // read 0..1023 from pin A0

        // when simulating adjust the sensorvalue to something kind of random, but smooth
        if (simulating) {
          sensorValue = simulate(inOutBreathingWindowInMsecs / 2); // hyperventilate
          //sensorValue = simulate(inOutBreathingWindowInMsecs); // awake
          //sensorValue = simulate(3 * inOutBreathingWindowInMsecs); // asleep
        }

        // save the value we just measured
        storeInCache(sensorValue);

        // TODO if we provide the logic in stead of the controlling Ardino
        //boolean isAsleep = computePersonSleeps();
        //if (isAsleep) {
        //  sendAlarm();
        //}

        // send result to the controller
        sendMeasurementResult(sensorValue);

        sendTops(getNumberOfHighs(topsWindowInSeconds));

        lastMeasuredValue = sensorValue;
      }
      break;
  }
  // wait for next cycle
  delay(loop_delay_msecs);        // delay in between reads for stability
}

int simulate(int breathingCycleTimeMsecs)
{
  int sensorValue = lastMeasuredValue;
  int stepSize = (int) (400.0 * (1000.0 / (float)breathingCycleTimeMsecs));

  if ((millis() % breathingCycleTimeMsecs) > (breathingCycleTimeMsecs / 2)) {
    sensorValue += random(stepSize); // breath in
    sensorValue = min(MAX_BELT_VALUE, sensorValue); // limit the value
  } else {
    sensorValue -= random(stepSize); // breath out
    sensorValue = max(0, sensorValue); // limit the value
  }

  return sensorValue;
}

boolean computePersonSleeps() {

  // The logic:

  // --- check breathing frequency
  int lastLowIndex = getLastLowIndex(inOutBreathingWindowInMsecs);
  int lastHighIndex = getLastHighIndex(inOutBreathingWindowInMsecs);
  // if max and min value elkaar snel opvolgen is iemand wakker
  // FIXME: dows not work at the end off the rrd cache!
  if (abs( lastLowIndex - lastHighIndex) > ASLEEP_FREQUENCY_THRESHOLD_IN_RRD_INDECES) {
    Serial.print("Asleep!");
    Serial.print(" Threshold=");
    Serial.print(ASLEEP_FREQUENCY_THRESHOLD_IN_RRD_INDECES);
    Serial.print("; Low=");
    Serial.print(lastLowIndex);
    Serial.print("; High=");
    Serial.println(lastHighIndex);

    return true;
  }

  // --- check trend flip
  currentTrend = getTrendReadingInLastMilliseconds(inOutBreathingWindowInMsecs);
  if (currentTrend != lastTrend) {
    waitingForTrendChange = 0;
    lastTrend = currentTrend;
  } else {
    waitingForTrendChange += loop_delay_msecs;
  }

  if (waitingForTrendChange > SLEEPING_THRESHOLD) {
    // we might be asleep!!! TODO
  }

  // compute min reading over last secs
  int minReading = getMinimumReadingInLastMilliseconds(inOutBreathingWindowInMsecs);
  // compute max reading over last secs
  int maxReading = getMaximumReadingInLastMilliseconds(inOutBreathingWindowInMsecs);
  // TODO add logic

  return false;
}

/**
   Store new value in the RRD.
*/
void storeInCache(int value) {
  cache[currentRRDIndex++] = value;

  if (currentRRDIndex > maxRRDsize) {
    currentRRDIndex = 0;
  }
}

/**
   Get the trend of the last reaadings.

   Returns
      -1 if the direction is downwards
      1  if the direction is upwards
      0 otherwise
*/
int getTrendReadingInLastMilliseconds(unsigned long mSecs) {
  int endIndex = getEndIndex();
  int startIndex = getStartIndex(mSecs);
  if (trace) printIndex(startIndex, endIndex);

  int up = 0;
  int down = 0;
  int lastValue = 0;
  int currentValue = 0;
  for (int k = startIndex; k < endIndex; k++) {
    currentValue = getFromCache(k);

    if (currentValue > lastValue) {
      up++;
    } else if (currentValue < lastValue) {
      down++;
    }
    lastValue = currentValue;
  }

  if (up > down) {
    return 1;
  } else if (down > up) {
    return -1;
  }

  return 0;
}

/**
   Get the index of the last low-value over the last msecs.
*/
int getLastLowIndex(unsigned long mSecs) {
  int endIndex = getEndIndex();
  int startIndex = getStartIndex(mSecs);
  if (trace) printIndex(startIndex, endIndex);

  int lowIndex = startIndex;
  int lastValue = 10000; // really big number
  for (int k = startIndex; k < endIndex; k++) {
    int indexValue = getFromCache(k);
    if (indexValue < lastValue) {
      lastValue = indexValue;
      lowIndex = k;
    }
  }
  return lowIndex % maxRRDsize;
}

/**
   Get the number of the high-values over the last number of seconds.
*/
const int LOOKING_FOR_TOP = 1;
const int LOOKING_FOR_BOTTOM = 2;

int getNumberOfHighs(unsigned long nrSecs) {
  int phase = LOOKING_FOR_TOP;
  int endIndex = getEndIndex();
  int startIndex = getStartIndexFromSeconds(nrSecs);
  if (trace) printIndex(startIndex, endIndex);

  int numberOfTops = 0;
  int numberOfBottoms = 0;
  int highIndex = startIndex; // front to back
  int highestValue = -10000; // really small number
  int lowestValue = 10000; // really big number
  for (int k = startIndex; k < endIndex; k++) {
    int cacheValue = getFromCache(k);
    if (traceTops) Serial.print(cacheValue);

    switch (phase) {
      case LOOKING_FOR_TOP:
        {
          if (cacheValue > highestValue) {
            highestValue = cacheValue;
            if (traceTops) Serial.print(":");
          }
          if (justPastTop(highestValue, cacheValue)) {
            lowestValue = 10000; // really big number
            numberOfTops++;
            if (traceTops) Serial.print("+");

            phase = LOOKING_FOR_BOTTOM;
          } else {
            if (traceTops) Serial.print(" ");
          }
          break;
        }
      case LOOKING_FOR_BOTTOM:
        {
          if (cacheValue < lowestValue) {
            lowestValue = cacheValue;
            if (traceTops) Serial.print("/");
          }
          if (justPastBottom(lowestValue, cacheValue)) {
            highestValue = -10000; // really small number
            numberOfBottoms++;
            if (traceTops) Serial.print("-");

            phase = LOOKING_FOR_TOP;
          } else {
            if (traceTops) Serial.print(" ");
          }
          break;
        }
    }
  }

  if (traceTops) Serial.println(".");
  return numberOfTops;
}

/**
    Check if the cacheValue is just over the top value measured
*/
boolean justPastTop(int highestValue, int cacheValue) {
  int difference = abs(highestValue - cacheValue);
  if ( (difference > (0.1 * maxBreathingRange)) && (difference < (0.25 * maxBreathingRange))) {
    return true;
  }
  return false;
}
/**
    Check if the cacheValue is just over the bottom value measured
*/
boolean justPastBottom(int lowestValue, int cacheValue) {
  int difference = abs(lowestValue - cacheValue);
  if ( (difference > (0.1 * maxBreathingRange)) && (difference < (0.25 * maxBreathingRange))) {
    return true;
  }
  return false;
}
/**
   Get the index of the last high-value over the last msecs.
*/
int getLastHighIndex(unsigned long mSecs) {
  int endIndex = getEndIndex();
  int startIndex = getStartIndex(mSecs);
  if (trace) printIndex(startIndex, endIndex);

  int highIndex = startIndex;
  int lastValue = -10000; // really small number
  for (int k = startIndex; k < endIndex; k++) {
    int indexValue = getFromCache(k);
    if (indexValue > lastValue) {
      lastValue = indexValue;
      highIndex = k;
    }
  }
  return highIndex % maxRRDsize;
}

/**
   Get the minimum value stored over the last msecs.
*/
int getMinimumReadingInLastMilliseconds(unsigned long mSecs) {
  int endIndex = getEndIndex();
  int startIndex = getStartIndex(mSecs);
  if (trace) printIndex(startIndex, endIndex);

  int minValue = 10000; // really big number
  for (int k = startIndex; k < endIndex; k++) {
    minValue = min(minValue, getFromCache(k));
  }
  return minValue;
}

/**
   Get the maximum value stored over the last msecs.
*/
int getMaximumReadingInLastMilliseconds(unsigned long mSecs) {
  int endIndex = getEndIndex();
  int startIndex = getStartIndex(mSecs);
  if (trace) printIndex(startIndex, endIndex);

  int maxValue = -10000; // really small number
  for (int k = startIndex; k < endIndex; k++) {
    maxValue = max(maxValue, getFromCache(k));
  }
  return maxValue;
}

/**
   Get the minimum value in the RRD.
*/
int getMinimumReadingMeasured() {
  int endIndex = maxRRDsize; // to avoid getting below 0 we increment with one RRD size.
  int startIndex = 0;
  int minValue = 10000; // really big number
  for (int k = startIndex; k < endIndex; k++) {
    minValue = min(minValue, getFromCache(k));
  }
  return minValue;
}

/**
    Get the maximum value in the RRD.
*/
int getMaximumReadingMeasured() {
  int endIndex = maxRRDsize; // to avoid getting below 0 we increment with one RRD size.
  int startIndex = 0;
  int maxValue = -10000; // really big number
  for (int k = startIndex; k < endIndex; k++) {
    maxValue = max(maxValue, getFromCache(k));
  }
  return maxValue;
}

/**
   Debugging output
*/
void printIndex(int s, int e)
{
  Serial.print(s);
  Serial.print("-");
  Serial.println(e);
}
/*********************************************************************************************************
   Functions to send data to the controller Arduino
*/

// send number of tops measured
void sendTops(int nrTops) {
  int now = millis() / 1000;
  Serial.print("(");
  Serial.print(now);
  Serial.print(") #Tops: ");
  Serial.println(nrTops);
}

// send measurement results
void sendMeasurementResult(int maxReading) {
  // TODO
  Serial.print("--> sending: ");
  Serial.println(maxReading);
}

// send an alarm
void sendAlarm() {
  // TODO
  Serial.println("!!! Alarm !!!");
}

/*********************************************************************************************************
   Functions to manipulate the RRD cache
*/

int getEndIndex()
{
  return ((currentRRDIndex - 1) + maxRRDsize); // to avoid getting below 0 we increment with one RRD size.
}

int getStartIndex(unsigned long backInHistoryInMsecs)
{
  return ((currentRRDIndex - 1) - (backInHistoryInMsecs / loop_delay_msecs) + maxRRDsize);
}

// sometimes the number of msecs is > 64K. Then we can use this function.
int getStartIndexFromSeconds(unsigned long backInHistoryInSecs)
{
  return ((currentRRDIndex - 1) - ( (float)backInHistoryInSecs / ( 1000.0 / (float)loop_delay_msecs )) + maxRRDsize);
}

int getFromCache(int k)
{
  return cache[k % maxRRDsize];
}
